package cz.fidentis.analyst.comparison;

import cz.fidentis.analyst.mesh.core.MeshFacet;
import cz.fidentis.analyst.mesh.io.MeshObjLoader;

import java.io.File;
import java.io.IOException;
import java.util.List;
import java.util.concurrent.CompletableFuture;

/**
 * @author Matej Lukes
 */
public class Comparison {
    private HausdorffDistance hausdorffDistance;

    private MeshFacet mainFacet;
    private MeshFacet comparedFacet;

    /**
     * Asynchronously loads main meshModel
     *
     * @param path path to meshModel file
     * @return CompletableFuture
     */
    public CompletableFuture loadMainModel(String path) {

        return CompletableFuture.runAsync(() -> {
            try {
                mainFacet = MeshObjLoader.read(new File(path)).getFacets().get(1);
            } catch (IOException e) {
                e.printStackTrace();
            }
        });
    }

    /**
     * Asynchronously loads compared meshModel
     *
     * @param path path to meshModel file
     * @return CompletableFuture
     */
    public CompletableFuture loadComparedModel(String path) {
        return CompletableFuture.runAsync(() -> {
            try {
                comparedFacet = MeshObjLoader.read(new File(path)).getFacets().get(1);
            } catch (IOException e) {
                e.printStackTrace();
            }
        });
    }

    /**
     * Asynchronously registers compared meshFacet to main meshFacet
     *
     * @param method registration method
     * @return CompletableFuture
     */
    public CompletableFuture register(RegistrationMethod method) {
        return CompletableFuture.runAsync(() -> comparedFacet = Registration
                .register(mainFacet, comparedFacet, method));
    }

    /**
     * Asynchronously compares MeshFacets from vertices to vertices
     *
     * @return list containing vertex from first facet, closest vertex to it from second facet, distance
     */
    public CompletableFuture<List<ClosestVertices>> compareHausdorffDistanceToVertices() {
        hausdorffDistance = new HausdorffDistance(mainFacet, comparedFacet);
        return CompletableFuture.supplyAsync(() -> hausdorffDistance.calculateHausdorffDistanceToVertices());
    }

    /**
     * Asynchronously compares MeshFacets from vertices to any point on mesh
     *
     * @return list containing vertex from first facet, closest point to it from second facet, distance
     */
    public CompletableFuture<List<ClosestVertices>> compareHausdorffDistanceToMesh() {
        hausdorffDistance = new HausdorffDistance(mainFacet, comparedFacet);
        return CompletableFuture.supplyAsync(() -> hausdorffDistance.calculateHausdorffDistanceToMesh());
    }

    /**
     * returns progress percentage
     *
     * @return progress percentage
     */
    public double getComparisonProgress() {
        if (hausdorffDistance == null) {
            return -1;
        }
        return hausdorffDistance.getProgressPercentage();
    }
}
