package cz.fidentis.analyst.mesh.KdTree;

import cz.fidentis.analyst.mesh.core.MeshPoint;


import javax.vecmath.Vector3d;
import java.util.*;

public class KdTreeImpl {
    private final KdNode root;
    private final List<MeshPoint> points;

    //private Map<MeshPoint, Integer> pointsWithIndex = new HashMap<>();

    private final static int X_AXIS = 0;
    private final static int Y_AXIS = 1;
    private final static int Z_AXIS = 2;




    public KdTreeImpl(List<MeshPoint> points) {
        if(points == null || points.isEmpty()){
            this.root = null;
            this.points = null;
            return;
        }

        this.points = points;

        List<Integer> pointsCopy = new ArrayList<>(points.size());
        for(int i = 0; i < points.size(); i++){
            pointsCopy.add(i);
        }

        pointsCopy = deleteDuplicates(points); //(pointsCopy,

        List<Integer> sortedByX = sortPoints(pointsCopy, X_AXIS, points);
        List<Integer> sortedByY = sortPoints(pointsCopy, Y_AXIS, points);
        List<Integer> sortedByZ = sortPoints(pointsCopy, Z_AXIS, points);


        root = buildTree(null, sortedByX, sortedByY, sortedByZ, 0, points);

    }

    private boolean comparePointsOnLevel(MeshPoint p1, MeshPoint p2, int level){
        if(level % 3 == 0){
            return p1.getPosition().x <= p2.getPosition().x;
        }else if(level % 3 == 1){
            return p1.getPosition().y <= p2.getPosition().y;
        }else if(level % 3 == 2){
            return p1.getPosition().z <= p2.getPosition().z;
        }

        return false;
    }

    private List<Integer> merge(List<Integer> left, List<Integer> right, int level, List<MeshPoint> points){
        List<Integer> mergedList = new ArrayList<>(left.size() + right.size());

        int fromLeft = 0;
        int fromRight = 0;

        while(fromLeft < left.size() || fromRight < right.size()){
            if(fromLeft < left.size() && fromRight < right.size()){
                if(comparePointsOnLevel(points.get(left.get(fromLeft)), points.get(right.get(fromRight)), level)){
                    mergedList.add(left.get(fromLeft));
                    fromLeft++;
                }else{
                    mergedList.add(right.get(fromRight));
                    fromRight++;
                }
            }else if(fromLeft < left.size()){
                mergedList.add(left.get(fromLeft));
                fromLeft++;
            }else if(fromRight < right.size()){
                mergedList.add(right.get(fromRight));
                fromRight++;
            }
        }

        return mergedList;
    }

    private List<Integer> mergeSort(List<Integer> points, int level, List<MeshPoint> p){
        if(points.size() <= 1){
            return points;
        }

        List<Integer> left;
        List<Integer> right;

        int mid = (points.size()/2);

        left = points.subList(0, mid);
        right = points.subList(mid, points.size());


        left = mergeSort(left, level, p);
        right = mergeSort(right, level, p);

        return merge(left, right, level, p);
    }

    private List<Integer> sortPoints(List<Integer> points, int level, List<MeshPoint> p){
        List<Integer> sortedList = new ArrayList<>(points.size());

        sortedList.addAll(points);

        sortedList = mergeSort(sortedList,level, p);

        return sortedList;

    }


    private List<Integer> deleteDuplicates(List<MeshPoint> points){ // List<Integer> list
        List<Vector3d> noDuplicates = new ArrayList<>();
        List<Integer> noDupIndex = new ArrayList<>();

        for(int i = 0; i < points.size(); i++){
            if(!noDuplicates.contains(points.get(i).getPosition())){
                noDuplicates.add(points.get(i).getPosition());
                noDupIndex.add(i);
            }
        }

        return noDupIndex;
    }


    private KdNode buildTree(KdNode parent, List<Integer> byX, List<Integer> byY, List<Integer> byZ, int level, List<MeshPoint> points) {
        KdNode node = null;
        int mid;
        int midIndex;

        if(byX.size() > 0 && byY.size() > 0 && byZ.size() > 0){
            mid = (byX.size() / 2);

            List<Integer> leftX = new ArrayList<>(mid);
            List<Integer> leftY = new ArrayList<>(mid);
            List<Integer> leftZ = new ArrayList<>(mid);

            //split lists in half, set middle element as new KdNode to be returned
            //first list to be split based on level, rest so that they contain points in first list
            //but also keep the ordering by their axis
            if (level % 3 == 0){
                midIndex = byX.get(mid);
                node = new KdNode(midIndex, level, points.get(byX.get(mid)), parent);

                splitTree(mid, leftX, byX, leftY, byY, leftZ, byZ);

            }else if(level % 3 == 1){
                midIndex = byY.get(mid);
                node = new KdNode(midIndex, level, points.get(byY.get(mid)), parent);

                splitTree(mid, leftY, byY, leftX, byX, leftZ, byZ);

            }else{
                midIndex = byZ.get(mid);
                node = new KdNode(midIndex, level, points.get(byZ.get(mid)), parent);

                splitTree(mid, leftZ, byZ, leftY, byY, leftX, byX);

            }

            //removes current middle node from each all lists, in case there were duplicates
            byX.removeAll(Collections.singleton(midIndex));
            byY.removeAll(Collections.singleton(midIndex));
            byZ.removeAll(Collections.singleton(midIndex));
            leftX.removeAll(Collections.singleton(midIndex));
            leftY.removeAll(Collections.singleton(midIndex));
            leftZ.removeAll(Collections.singleton(midIndex));

            node.setLesser(buildTree(node, leftX, leftY, leftZ, level + 1, points));
            node.setGreater(buildTree(node, byX, byY, byZ, level + 1, points));
        }

        return node;
    }

    private void splitTree(int mid, List<Integer> leftMain, List<Integer> mainList, List<Integer> leftSecond,
                           List<Integer> secondList, List<Integer> leftThird, List<Integer> thirdList){
        for (int i = 0; i < mid; i++) {

            leftMain.add(mainList.get(0));

            mainList.remove(0);
        }

        splitArray(leftMain, leftSecond, secondList);
        splitArray(leftMain, leftThird, thirdList);
    }

    private void splitArray(List<Integer> leftMain, List<Integer> leftSecond, List<Integer> secondList){
        for(int i = secondList.size() - 1; i >= 0; i--){
            if(leftMain.contains(secondList.get(i))){
                leftSecond.add(0, secondList.get(i));
                secondList.remove(i);
            }
        }
    }
    public MeshPoint nearestNeighbour(MeshPoint p){
        KdNode nn = nearestNeighbourNode(p);

        if(nn == null){
            return null;
        }

        return nn.getId();
    }

    public boolean containPoint(MeshPoint p){
        MeshPoint found = nearestNeighbour(p);

        if(found == null){
            return false;
        }

        return found.equals(p);
    }

     public double distancePoints(MeshPoint point1, MeshPoint point2) {
        return Math.sqrt(Math.pow(point1.getPosition().x - point2.getPosition().x, 2)
                + Math.pow(point1.getPosition().y - point2.getPosition().y, 2)
                + Math.pow(point1.getPosition().z - point2.getPosition().z, 2));
    }

    private double minDistanceIntersection(MeshPoint searched, MeshPoint p2, int level){
        if(level % 3 == 0){
            return Math.abs(searched.getPosition().x - p2.getPosition().x);
        }else if(level % 3 == 1){
            return Math.abs(searched.getPosition().y - p2.getPosition().y);
        }else{
            return Math.abs(searched.getPosition().z - p2.getPosition().z);
        }
    }



    public KdNode nearestNeighbourNode(MeshPoint p) {
        if(p == null || this.root == null){
            return null;
        }

        double minDistance = Double.MAX_VALUE;
        double distOnAxis;
        KdNode near = root;
        KdNode searched = root;

        Queue<KdNode> queue = new LinkedList<>();

        while(searched != null){
            if(distancePoints(p, searched.getId()) < minDistance){
                near = searched;
                minDistance = distancePoints(p, searched.getId());
            }

            if(comparePointsOnLevel(p, searched.getId(), searched.getDepth())){
                searched = searched.getLesser();
            }else{
                searched = searched.getGreater();
            }
        }

        queue.add(root);

        //second search to find vertex that could be potentially closer than
        //nearest vertex already found
        while (!queue.isEmpty()) {

            //if min distance is 0, break cycle (nothing can be closer)
            if(minDistance == 0){
                break;
            }

            searched = queue.poll();

            if (distancePoints(p, searched.getId()) < minDistance) {
                near = searched;
                minDistance = distancePoints(p, searched.getId());
            }

            distOnAxis = minDistanceIntersection(searched.getId(), p, searched.getDepth());

            if (distOnAxis > minDistance) {
                if (comparePointsOnLevel(p, searched.getId(), searched.getDepth())) {
                    if (searched.getLesser() != null) {
                        queue.add(searched.getLesser());
                    }
                } else {
                    if (searched.getGreater() != null) {
                        queue.add(searched.getGreater());
                    }
                }
            } else {
                if (searched.getLesser() != null) {
                    queue.add(searched.getLesser());
                }
                if (searched.getGreater() != null) {
                    queue.add(searched.getGreater());
                }
            }
        }

        return near;

    }

    /**
     * Find index of MeshPoint in MeshFacetImpl -> vertices
     * @param p MeshPoint
     * @return index of MeshPoint in MeshFacetImpl -> vertices
     */
    public Integer nearestNeighbourIndex(MeshPoint p){
        KdNode nn = nearestNeighbourNode(p);

        if(nn == null){
            return null;
        }

        return nn.getIndex();
    }

}
